@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
- * Copyright (C) 2003-2005 Imendio AB
+ * Copyright (C) 2003-2006 Imendio AB
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License as
@@ -21,25 +21,21 @@
 #include <config.h>
 
 #include <string.h>
-#include <unistd.h>
+#include <sys/stat.h> 
 #include <sys/types.h>
 #include <fcntl.h>
-#include <errno.h>
-#ifndef __WIN32__
-  #include <netdb.h>
-  #include <sys/socket.h>
-  #include <netinet/in.h>
-  #include <sys/time.h>
-#else
-  #include <winsock2.h>
-#endif
 
+#include <glib.h>
+
+#include "lm-sock.h"
 #include "lm-debug.h"
 #include "lm-error.h"
 #include "lm-internals.h"
+#include "lm-ssl-internals.h"
 #include "lm-parser.h"
 #include "lm-sha.h"
 #include "lm-connection.h"
+#include "lm-utils.h"
 
 #define IN_BUFFER_SIZE 1024
 
@@ -75,12 +71,15 @@ struct _LmConnection {
 	guint         io_watch_in;
 	guint         io_watch_err;
 	guint         io_watch_hup;
-	guint         fd;
+	LmSocket      fd;
 	guint         io_watch_connect;
 	
 	guint         open_id;
 	LmCallback   *open_cb;
 
+ 	gboolean      async_connect_waiting;
+	gboolean      blocking;
+
 	gboolean      cancel_open;
 	LmCallback   *close_cb;     /* unused */
 	LmCallback   *auth_cb;
@@ -99,6 +98,8 @@ struct _LmConnection {
 	guint         io_watch_out;
 	GString      *out_buf;
 
+	LmConnectData *connect_data;
+
 	gint          ref_count;
 };
 
@@ -204,6 +205,7 @@ connection_free (LmConnection *connection)
 
 	g_free (connection->server);
 	g_free (connection->jid);
+	g_free (connection->stream_id);
 
 	if (connection->parser) {
 		lm_parser_free (connection->parser);
@@ -246,6 +248,10 @@ connection_free (LmConnection *connection)
 		lm_ssl_unref (connection->ssl);
 	}
 
+	if (connection->proxy) {
+		lm_proxy_unref (connection->proxy);
+	}
+
 	g_queue_free (connection->incoming_messages);
         
         if (connection->context) {
@@ -327,97 +333,122 @@ connection_new_message_cb (LmParser     *parser,
 	g_queue_push_tail (connection->incoming_messages, m);
 }
 
-gboolean
+void
 _lm_connection_succeeded (LmConnectData *connect_data)
 {
 	LmConnection *connection;
 	LmMessage    *m;
-	GIOFlags      flags;
 	gchar        *server_from_jid;
 	gchar        *ch;
+	gchar        *ch_end;
 
 	connection = connect_data->connection;
 	
 	if (connection->io_watch_connect != 0) {
-		g_source_destroy (g_main_context_find_source_by_id(connection->context,
-								   connection->io_watch_connect));
+		GSource *source;
+
+		source = g_main_context_find_source_by_id (connection->context,
+							   connection->io_watch_connect);
+		if (source) {
+			g_source_destroy (source);
+		}
 		connection->io_watch_connect = 0;
 	}
 
 	/* Need some way to report error/success */
 	if (connection->cancel_open) {
-		return FALSE;
+		lm_verbose ("Cancelling connection...\n");
+		return;
 	}
 	
 	connection->fd = connect_data->fd;
 	connection->io_channel = connect_data->io_channel;
 
 	freeaddrinfo (connect_data->resolved_addrs);
-
-	/* don't need this anymore */
-	g_free(connect_data);
-
-	flags = g_io_channel_get_flags (connection->io_channel);
-
-	/* unset the nonblock flag */
-	flags &= ~G_IO_FLAG_NONBLOCK;
-	
-	/* unset the nonblocking stuff for some time, because GNUTLS doesn't 
-	 * like that */
-	g_io_channel_set_flags (connection->io_channel, flags, NULL);
+	connection->connect_data = NULL;
+	g_free (connect_data);
 
 	if (connection->ssl) {
+		GError *error = NULL;
+
+		lm_verbose ("Setting up SSL...\n");
+	
+#ifdef HAVE_GNUTLS
+		/* GNU TLS requires the socket to be blocking */
+		_lm_sock_set_blocking (connection->fd, TRUE);
+#endif
+
 		if (!_lm_ssl_begin (connection->ssl, connection->fd,
 				    connection->server,
-				    NULL)) {
-			shutdown (connection->fd, SHUT_RDWR);
-			close (connection->fd);
-		
+				    &error)) {
+			lm_verbose ("Could not begin SSL\n");
+				    
+			if (error) {
+				g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET,
+				       "%s\n", error->message);
+				g_error_free (error);
+			}
+
+ 			_lm_sock_shutdown (connection->fd);
+ 			_lm_sock_close (connection->fd);
+
 			connection_do_close (connection);
-			return FALSE;
+
+			return;
 		}
+	
+#ifdef HAVE_GNUTLS
+		_lm_sock_set_blocking (connection->fd, FALSE); 
+#endif
 	}
-	
-	g_io_channel_set_close_on_unref (connection->io_channel, TRUE);
-	g_io_channel_set_encoding (connection->io_channel, NULL, NULL);
-	
-	g_io_channel_set_buffered (connection->io_channel, FALSE);
-	g_io_channel_set_flags (connection->io_channel,
-				flags & G_IO_FLAG_NONBLOCK, NULL);
-	
+
 	connection->io_watch_in = 
 		connection_add_watch (connection,
 				      connection->io_channel,
 				      G_IO_IN,
 				      (GIOFunc) connection_in_event,
 				      connection);
-	
-	connection->io_watch_err = 
-		connection_add_watch (connection,
-				      connection->io_channel, 
-				      G_IO_ERR,
-				      (GIOFunc) connection_error_event,
-						   connection);
-	connection->io_watch_hup = 
-		connection_add_watch (connection,
-				      connection->io_channel,
-				      G_IO_HUP,
-				      (GIOFunc) connection_hup_event,
-				      connection);
+
+	/* FIXME: if we add these, we don't get ANY
+	 * response from the server, this is to do with the way that
+	 * windows handles watches, see bug #331214.
+	 */
+#ifndef G_OS_WIN32
+		connection->io_watch_err =
+			connection_add_watch (connection,
+					      connection->io_channel,
+					      G_IO_ERR,
+					      (GIOFunc) connection_error_event,
+					      connection);
+		
+		connection->io_watch_hup =
+			connection_add_watch (connection,
+					      connection->io_channel,
+					      G_IO_HUP,
+					      (GIOFunc) connection_hup_event,
+					      connection);
+#endif
 
 	/* FIXME: Set up according to XMPP 1.0 specification */
 	/*        StartTLS and the like */
 	if (!connection_send (connection, 
 			      "<?xml version='1.0' encoding='UTF-8'?>", -1,
 			      NULL)) {
+		lm_verbose ("Failed to send xml version and encoding\n");
 		connection_do_close (connection);
-		return FALSE;
+
+		return;
 	}
 
 	if (connection->jid != NULL && (ch = strchr (connection->jid, '@')) != NULL) {
-		server_from_jid = ch + 1;
+		ch_end = strchr(ch + 1, '/');
+		if (ch_end != NULL) {
+			server_from_jid = g_strndup (ch + 1, ch_end - ch - 1);
+		} else {
+			server_from_jid = g_strdup (ch + 1);
+		}
 	} else {
-		server_from_jid = connection->server;
+		server_from_jid = g_strdup (connection->server);
 	}
 
 	m = lm_message_new (server_from_jid, LM_MESSAGE_TYPE_STREAM);
@@ -428,17 +459,14 @@ _lm_connection_succeeded (LmConnectData *connect_data)
 					NULL);
 	
 	lm_verbose ("Opening stream...");
-	
+
 	if (!lm_connection_send (connection, m, NULL)) {
-		lm_message_unref (m);
+		lm_verbose ("Failed to send stream information\n");
 		connection_do_close (connection);
-		return FALSE;
 	}
 		
+	g_free (server_from_jid);
 	lm_message_unref (m);
-
-	/* Success */
-	return FALSE;
 }
 
 void 
@@ -448,28 +476,34 @@ _lm_connection_failed_with_error (LmConnectData *connect_data, int error)
 	
 	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET,
 	       "Connection failed: %s (error %d)\n",
-	       strerror (error), error);
+	       _lm_sock_get_error_str (error), error);
 	
 	connection = connect_data->connection;
 	
 	connect_data->current_addr = connect_data->current_addr->ai_next;
 	
 	if (connection->io_watch_connect != 0) {
-		g_source_destroy (g_main_context_find_source_by_id(connection->context,
-								   connection->io_watch_connect));
+		GSource *source;
+
+		source = g_main_context_find_source_by_id (connection->context,
+							   connection->io_watch_connect);
+		if (source) {
+			g_source_destroy (source);
+		}
+
 		connection->io_watch_connect = 0;
 	}
 
 	if (connect_data->io_channel != NULL) {
 		g_io_channel_unref (connect_data->io_channel);
+		/* FIXME: need to check for last unref and close the socket */
 	}
 	
 	if (connect_data->current_addr == NULL) {
 		connection_do_close (connection);
 		if (connection->open_cb) {
-			LmCallback *cb;
+			LmCallback *cb = connection->open_cb;
 
-			cb = connection->open_cb;
 			connection->open_cb = NULL;
 			
 			(* ((LmResultFunction) cb->func)) (connection, FALSE,
@@ -478,6 +512,7 @@ _lm_connection_failed_with_error (LmConnectData *connect_data, int error)
 		}
 		
 		freeaddrinfo (connect_data->resolved_addrs);
+		connection->connect_data = NULL;
 		g_free (connect_data);
 	} else {
 		/* try to connect to the next host */
@@ -488,43 +523,107 @@ _lm_connection_failed_with_error (LmConnectData *connect_data, int error)
 void 
 _lm_connection_failed (LmConnectData *connect_data)
 {
-	_lm_connection_failed_with_error (connect_data,errno);
+	_lm_connection_failed_with_error (connect_data, 
+					  _lm_sock_get_last_error());
 }
 
 static gboolean 
 connection_connect_cb (GIOChannel   *source, 
 		       GIOCondition  condition,
-		       gpointer      data) 
+		       LmConnectData *connect_data) 
 {
-	LmConnectData *connect_data;
-	int            error;
-	guint          len  = sizeof(error);
+	LmConnection    *connection;
+	struct addrinfo *addr;
+	int              err;
+	socklen_t        len;
+	LmSocket         fd; 
 
-	connect_data = (LmConnectData *) data;
+	connection = connect_data->connection;
+	addr = connect_data->current_addr;
+	fd = g_io_channel_unix_get_fd (source);
 	
 	if (condition == G_IO_ERR) {
-		/* get the real error from the socket */
-		getsockopt (connect_data->fd, SOL_SOCKET, SO_ERROR, 
-			    &error, &len);
-		_lm_connection_failed_with_error (connect_data, error);
-		return FALSE;
-	} else if (condition == G_IO_OUT) {
-		_lm_connection_succeeded (connect_data);
-	} else {
-		g_assert_not_reached ();
+		len = sizeof (err);
+		_lm_sock_get_error (fd, &err, &len);
+		if (!_lm_sock_is_blocking_error (err)) {
+			g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET,
+			       "Connection failed.\n");
+
+			_lm_connection_failed_with_error (connect_data, err);
+
+			connection->io_watch_connect = 0;
+			return FALSE;
+		}
 	}
 
-	return FALSE;
+	if (connection->async_connect_waiting) {
+		gint res;
+
+		fd = g_io_channel_unix_get_fd (source);
+
+		res = _lm_sock_connect (fd, addr->ai_addr, (int)addr->ai_addrlen);  
+		if (res < 0) {
+			err = _lm_sock_get_last_error ();
+			if (_lm_sock_is_blocking_success (err)) {
+				connection->async_connect_waiting = FALSE;
+
+				g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET,
+				       "Connection success.\n");
+				
+				_lm_connection_succeeded (connect_data);
+			}
+			
+			if (connection->async_connect_waiting && 
+			    !_lm_sock_is_blocking_error (err)) {
+				g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET,
+				       "Connection failed.\n");
+
+				_lm_sock_close (connect_data->fd);
+				_lm_connection_failed_with_error (connect_data, err);
+
+				connection->io_watch_connect = 0;
+				return FALSE;
+			}
+		} 
+	} else {		
+		/* for blocking sockets, G_IO_OUT means we are connected */
+		g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET,
+		       "Connection success.\n");
+		
+		_lm_connection_succeeded (connect_data);
+	}
+
+ 	return TRUE; 
+}
+
+static const char *
+connection_condition_to_str (GIOCondition condition)
+{
+	static char buf[256];
+
+	buf[0] = '\0';
+
+	if(condition & G_IO_ERR)
+		strcat(buf, "G_IO_ERR ");
+	if(condition & G_IO_HUP)
+		strcat(buf, "G_IO_HUP ");
+	if(condition & G_IO_NVAL)
+		strcat(buf, "G_IO_NVAL ");
+	if(condition & G_IO_IN)
+		strcat(buf, "G_IO_IN ");
+	if(condition & G_IO_OUT)
+		strcat(buf, "G_IO_OUT ");
+
+	return buf;
 }
 
 static void
 connection_do_connect (LmConnectData *connect_data) 
 {
 	LmConnection    *connection;
-	int              fd;
-	int              res;
+	LmSocket         fd;
+	int              res, err;
 	int              port;
-	int              flags;
 	char             name[NI_MAXHOST];
 	char             portname[NI_MAXSERV];
 	struct addrinfo *addr;
@@ -540,37 +639,47 @@ connection_do_connect (LmConnectData *connect_data)
 	
 	((struct sockaddr_in *) addr->ai_addr)->sin_port = port;
 
-	getnameinfo (addr->ai_addr,
-		     addr->ai_addrlen,
-		     name,     sizeof (name),
-		     portname, sizeof (portname),
-		     NI_NUMERICHOST | NI_NUMERICSERV);
+	res = getnameinfo (addr->ai_addr,
+			   (socklen_t)addr->ai_addrlen,
+			   name,     sizeof (name),
+			   portname, sizeof (portname),
+			   NI_NUMERICHOST | NI_NUMERICSERV);
+	
+	if (res < 0) {
+		_lm_connection_failed (connect_data);
+		return;
+	}
 
-	g_log (LM_LOG_DOMAIN,LM_LOG_LEVEL_NET,
+	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET,
 	       "Trying %s port %s...\n", name, portname);
 	
-	fd = socket (addr->ai_family, 
-		     addr->ai_socktype, 
-		     addr->ai_protocol);
+	fd = _lm_sock_makesocket (addr->ai_family,
+				  addr->ai_socktype, 
+				  addr->ai_protocol);
 	
-	if (fd < 0) {
+	if (!_LM_SOCK_VALID (fd)) {
+		g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, 
+		       "Failed making socket, error:%d...\n",
+		       _lm_sock_get_last_error ());
+
 		_lm_connection_failed (connect_data);
+
 		return;
 	}
 
-	flags = fcntl (fd, F_GETFL, 0);
-	fcntl (fd, F_SETFL, flags | O_NONBLOCK);
-	
-	res = connect (fd, addr->ai_addr, addr->ai_addrlen);
+	/* Even though it says _unix_new(), it is supported by glib on
+	 * win32 because glib does some cool stuff to find out if it
+	 * can treat it as a FD or a windows SOCKET.
+	 */
 	connect_data->fd = fd;
-	
-	if (res < 0 && errno != EINPROGRESS) {
-		close (fd);
-		_lm_connection_failed (connect_data);
-		return;
-	}
-	
 	connect_data->io_channel = g_io_channel_unix_new (fd);
+	
+	g_io_channel_set_encoding (connect_data->io_channel, NULL, NULL);
+	g_io_channel_set_buffered (connect_data->io_channel, FALSE);
+
+	_lm_sock_set_blocking (connect_data->fd, 
+			       connection->blocking);
+	
 	if (connection->proxy) {
 		connection->io_watch_connect =
 		connection_add_watch (connection,
@@ -587,7 +696,19 @@ connection_do_connect (LmConnectData *connect_data)
 				      connect_data);
 	}
 
-	return;
+	connection->async_connect_waiting = !connection->blocking;
+
+  	res = _lm_sock_connect (connect_data->fd, 
+				addr->ai_addr, (int)addr->ai_addrlen);  
+	if (res < 0) {
+		err = _lm_sock_get_last_error ();
+		if (!_lm_sock_is_blocking_error (err)) {
+			_lm_sock_close (connect_data->fd);
+			_lm_connection_failed_with_error (connect_data, err);
+
+			return;
+		}
+	}
 }
 
 static guint
@@ -617,7 +738,7 @@ static gboolean
 connection_send_keep_alive (LmConnection *connection)
 { 
 	if (!connection_send (connection, " ", -1, NULL)) {
-		lm_verbose ("Error while sending keep alive package");
+		lm_verbose ("Error while sending keep alive package!\n");
 	}
 
 	return TRUE;
@@ -675,9 +796,19 @@ connection_buffered_write_cb (GIOChannel   *source,
 	g_string_erase (out_buf, 0, (gsize) b_written);
 	if (out_buf->len == 0) {
 		lm_verbose ("Output buffer is empty, going back to normal output\n");
-		g_source_destroy (g_main_context_find_source_by_id (
-			connection->context, connection->io_watch_out));
-		connection->io_watch_out = 0;
+
+		if (connection->io_watch_out != 0) {
+			GSource *source;
+
+			source = g_main_context_find_source_by_id (connection->context, 
+								   connection->io_watch_out);
+			if (source) {
+				g_source_destroy (source);
+			}
+
+			connection->io_watch_out = 0;
+		}
+
 		g_string_free (out_buf, TRUE);
 		connection->out_buf = NULL;
 		return FALSE;
@@ -729,7 +860,7 @@ connection_do_open (LmConnection *connection, GError **error)
 	if (lm_connection_is_open (connection)) {
 		g_set_error (error,
 			     LM_ERROR,
-			     LM_ERROR_CONNECTION_OPEN,
+			     LM_ERROR_CONNECTION_NOT_OPEN,
 			     "Connection is already open, call lm_connection_close() first");
 		return FALSE;
 	}
@@ -757,33 +888,47 @@ connection_do_open (LmConnection *connection, GError **error)
 	
 	connection->cancel_open = FALSE;
 	connection->state = LM_CONNECTION_STATE_OPENING;
+	connection->async_connect_waiting = FALSE;
 	
 	if (connection->proxy) {
+		int          err;
 		const gchar *proxy_server;
 
 		proxy_server = lm_proxy_get_server (connection->proxy);
-		/* connect through proxy */
-		g_log (LM_LOG_DOMAIN,LM_LOG_LEVEL_NET,
-		       "Going to connect to %s\n", proxy_server);
 
-		if (getaddrinfo (proxy_server, NULL, &req, &ans) != 0) {
+		/* Connect through proxy */
+		g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET,
+		       "Going to connect to proxy %s\n", proxy_server);
+
+		err = getaddrinfo (proxy_server, NULL, &req, &ans);
+		if (err != 0) {
+			const char *str;
+
+			str = _lm_sock_addrinfo_get_error_str (err);
 			g_set_error (error,
 				     LM_ERROR,                 
 				     LM_ERROR_CONNECTION_FAILED,   
-				     "getaddrinfo() failed");
+				     str);
 			return FALSE;
 		}
-	} else { /* connect directly */
-		g_log (LM_LOG_DOMAIN,LM_LOG_LEVEL_NET,
+	} else { 
+		int err;
+
+		/* Connect directly */
+		g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET,
 		       "Going to connect to %s\n", 
 		       connection->server);
 
-		if (getaddrinfo (connection->server,
-				 NULL, &req, &ans) != 0) {
+		err = getaddrinfo (connection->server, 
+				   NULL, &req, &ans);
+		if (err != 0) {
+			const char *str;
+
+			str = _lm_sock_addrinfo_get_error_str (err);
 			g_set_error (error,
 				     LM_ERROR,                 
 				     LM_ERROR_CONNECTION_FAILED,   
-				     "getaddrinfo() failed");
+				     str);
 			return FALSE;
 		}
 	}
@@ -801,6 +946,8 @@ connection_do_open (LmConnection *connection, GError **error)
 	data->io_channel     = NULL;
 	data->fd             = -1;
 
+	connection->connect_data = data;
+
 	connection_do_connect (data);
 	return TRUE;
 }
@@ -808,44 +955,95 @@ connection_do_open (LmConnection *connection, GError **error)
 static void
 connection_do_close (LmConnection *connection)
 {
+	GSource       *source;
+	LmConnectData *data;
+
 	connection_stop_keep_alive (connection);
 
+	if (connection->io_watch_connect != 0) {
+
+		source = g_main_context_find_source_by_id (connection->context,
+							   connection->io_watch_connect);
+
+		if (source) {
+			g_source_destroy (source);
+		}
+
+		connection->io_watch_connect = 0;
+	}
+	
+	data = connection->connect_data;
+	if (data) {
+		freeaddrinfo (data->resolved_addrs);
+		connection->connect_data = NULL;
+		g_free (data);
+	}
+
 	if (connection->io_channel) {
-		g_source_destroy (g_main_context_find_source_by_id (
-			connection->context, connection->io_watch_in));
-		g_source_destroy (g_main_context_find_source_by_id (
-			connection->context, connection->io_watch_err));
-		g_source_destroy (g_main_context_find_source_by_id (
-			connection->context, connection->io_watch_hup));
+		if (connection->io_watch_in != 0) {
+			source = g_main_context_find_source_by_id (connection->context,
+								   connection->io_watch_in);
+			if (source) {
+				g_source_destroy (source);
+			}
+			
+			connection->io_watch_in = 0;
+		}
+
+		if (connection->io_watch_err != 0) {
+			source = g_main_context_find_source_by_id (connection->context, 
+								   connection->io_watch_err);
+			if (source) {
+				g_source_destroy (source);
+			}
+
+			connection->io_watch_err = 0;
+		}
+
+		if (connection->io_watch_hup != 0) {
+			source = g_main_context_find_source_by_id (connection->context, 
+								   connection->io_watch_hup);
+
+			if (source) {
+				g_source_destroy (source);
+			}
+			
+			connection->io_watch_hup = 0;
+		}
 
 		if (connection->io_watch_out != 0) {
-			g_source_destroy (g_main_context_find_source_by_id (
-				connection->context, connection->io_watch_out));
+			source = g_main_context_find_source_by_id (connection->context, 
+								   connection->io_watch_out);
+
+			if (source) {
+				g_source_destroy (source);
+			}
+
 			connection->io_watch_out = 0;
 		}
 
-		if (connection->io_watch_connect != 0) {
-			g_source_destroy (g_main_context_find_source_by_id(connection->context,
-									   connection->io_watch_connect));
-			connection->io_watch_connect = 0;
-		}
 
 		g_io_channel_unref (connection->io_channel);
 		connection->io_channel = NULL;
 
 		connection->fd = -1;
 	}
-	
-	g_source_destroy (connection->incoming_source);
-	g_source_unref (connection->incoming_source);
+
+	if (connection->incoming_source) {
+		g_source_destroy (connection->incoming_source);
+		g_source_unref (connection->incoming_source);
+		connection->incoming_source = NULL;
+	}
 
 	if (!lm_connection_is_open (connection)) {
 		/* lm_connection_is_open is FALSE for state OPENING as well */
 		connection->state = LM_CONNECTION_STATE_CLOSED;
+		connection->async_connect_waiting = FALSE;
 		return;
 	}
 	
 	connection->state = LM_CONNECTION_STATE_CLOSED;
+	connection->async_connect_waiting = FALSE;
 	if (connection->ssl) {
 		_lm_ssl_close (connection->ssl);
 	}
@@ -882,29 +1080,27 @@ connection_do_write (LmConnection *connection,
 }
 
 static gboolean
-connection_in_event (GIOChannel   *source,
-		     GIOCondition  condition,
-		     LmConnection *connection)
+connection_read_incoming (LmConnection *connection,
+			  gchar        *buf,
+			  gsize         buf_size,
+			  gsize        *bytes_read,
+			  gboolean     *hangup)
 {
-	gchar     buf[IN_BUFFER_SIZE];
-	gsize     bytes_read;
 	GIOStatus status;
-       
-	if (!connection->io_channel) {
-		return FALSE;
-	}
+
+	*hangup = FALSE;
 
 	if (connection->ssl) {
 		status = _lm_ssl_read (connection->ssl, 
-				       buf, IN_BUFFER_SIZE - 1, &bytes_read);
+				       buf, buf_size - 1, bytes_read);
 	} else {
 		status = g_io_channel_read_chars (connection->io_channel,
-						  buf, IN_BUFFER_SIZE - 1,
-						  &bytes_read,
+						  buf, buf_size - 1,
+						  bytes_read,
 						  NULL);
 	}
 
-	if (status != G_IO_STATUS_NORMAL || bytes_read < 0) {
+	if (status != G_IO_STATUS_NORMAL || *bytes_read < 0) {
 		gint reason;
 		
 		switch (status) {
@@ -912,7 +1108,8 @@ connection_in_event (GIOChannel   *source,
 			reason = LM_DISCONNECT_REASON_HUP;
 			break;
 		case G_IO_STATUS_AGAIN:
-			return TRUE;
+			/* No data readable but we didn't hangup */
+			return FALSE;
 			break;
 		case G_IO_STATUS_ERROR:
 			reason = LM_DISCONNECT_REASON_ERROR;
@@ -923,23 +1120,51 @@ connection_in_event (GIOChannel   *source,
 
 		connection_do_close (connection);
 		connection_signal_disconnect (connection, reason);
+
+		/* Notify connection_in_event that we hangup the connection */
+		*hangup = TRUE;
 		
 		return FALSE;
 	}
 
-	buf[bytes_read] = '\0';
-	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, "\nRECV [%d]:\n", 
-	       (int)bytes_read);
-	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, 
-	       "-----------------------------------\n");
-	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, "'%s'\n", buf);
-	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, 
-	       "-----------------------------------\n");
+	buf[*bytes_read] = '\0';
 
-	lm_verbose ("Read: %d chars\n", (int)bytes_read);
+	/* There is more data to be read */
+	return TRUE;
+}
+
+static gboolean
+connection_in_event (GIOChannel   *source,
+		     GIOCondition  condition,
+		     LmConnection *connection)
+{
+	gchar    buf[IN_BUFFER_SIZE];
+	gsize    bytes_read;
+	gboolean hangup;
+
+	if (connection->io_channel == NULL) {
+		return FALSE;
+	}
+
+	while (connection_read_incoming (connection, buf, IN_BUFFER_SIZE,
+					 &bytes_read, &hangup)) {
+		g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, "\nRECV [%d]:\n", 
+		       (int)bytes_read);
+		g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, 
+		       "-----------------------------------\n");
+		g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, "'%s'\n", buf);
+		g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, 
+		       "-----------------------------------\n");
+
+		lm_verbose ("Read: %d chars\n", (int)bytes_read);
+
+		lm_parser_parse (connection->parser, buf);
+	}
+
+	if (hangup) {
+		return FALSE;
+	}
 
-	lm_parser_parse (connection->parser, buf);
-	
 	return TRUE;
 }
 
@@ -948,12 +1173,13 @@ connection_error_event (GIOChannel   *source,
 			GIOCondition  condition,
 			LmConnection *connection)
 {
+	lm_verbose ("Error event: %d->'%s'\n", 
+		    condition, connection_condition_to_str (condition));
+
 	if (!connection->io_channel) {
 		return FALSE;
 	}
 
-	lm_verbose ("Error event: %d\n", condition);
-	
 	connection_do_close (connection);
 	connection_signal_disconnect (connection, LM_DISCONNECT_REASON_ERROR);
 	
@@ -965,12 +1191,13 @@ connection_hup_event (GIOChannel   *source,
 		      GIOCondition  condition,
 		      LmConnection *connection)
 {
+	lm_verbose ("HUP event: %d->'%s'\n", 
+		    condition, connection_condition_to_str (condition));
+
 	if (!connection->io_channel) {
 		return FALSE;
 	}
 
-	lm_verbose ("HUP event\n");
-
 	connection_do_close (connection);
 	connection_signal_disconnect (connection, LM_DISCONNECT_REASON_HUP);
 	
@@ -986,6 +1213,9 @@ connection_send (LmConnection  *connection,
 	gint b_written;
 	
 	if (connection->state < LM_CONNECTION_STATE_OPENING) {
+		g_log (LM_LOG_DOMAIN,LM_LOG_LEVEL_NET,
+		       "Connection is not open.\n");
+
 		g_set_error (error,
 			     LM_ERROR,
 			     LM_ERROR_CONNECTION_NOT_OPEN,
@@ -1193,9 +1423,8 @@ connection_auth_reply (LmMessageHandler *handler,
 	lm_verbose ("AUTH reply: %d\n", result);
 	
 	if (connection->auth_cb) {
-	        LmCallback *cb;
+	        LmCallback *cb = connection->auth_cb;
 
-		cb = connection->auth_cb;
 		connection->auth_cb = NULL;
 
 		if (cb->func) {
@@ -1231,16 +1460,15 @@ connection_stream_received (LmConnection *connection, LmMessage *m)
 	connection_start_keep_alive (connection);
 
 	if (connection->open_cb) {
-		LmCallback *cb;
+		LmCallback *cb = connection->open_cb;
 
-		cb = connection->open_cb;
 		connection->open_cb = NULL;
 		
 		if (cb->func) {
 		        (* ((LmResultFunction) cb->func)) (connection, result,
 			        			   cb->user_data);
-		}
 
+		}
 		_lm_utils_free_callback (connection->open_cb);
 	}
 }
@@ -1327,11 +1555,12 @@ lm_connection_new (const gchar *server)
 	gint          i;
 	
 	lm_debug_init ();
-	
+	_lm_sock_library_init ();
+
 	connection = g_new0 (LmConnection, 1);
 
 	if (server) {
-		connection->server = g_strdup (server);
+		connection->server = _lm_utils_hostname_to_punycode (server);
 	} else {
 		connection->server = NULL;
 	}
@@ -1348,6 +1577,7 @@ lm_connection_new (const gchar *server)
 	connection->keep_alive_id     = 0;
 	connection->keep_alive_rate   = 0;
 	connection->out_buf           = NULL;
+	connection->connect_data      = NULL;
 	
 	connection->id_handlers = g_hash_table_new_full (g_str_hash, 
 							 g_str_equal,
@@ -1385,7 +1615,9 @@ lm_connection_new_with_context (const gchar *server, GMainContext *context)
 	connection = lm_connection_new (server);
 	connection->context = context;
 
-        g_main_context_ref (connection->context);
+	if (context) {
+        	g_main_context_ref (connection->context);
+	}
 
 	return connection;
 }
@@ -1413,6 +1645,7 @@ lm_connection_open (LmConnection      *connection,
 	
 	connection->open_cb = _lm_utils_new_callback (function, 
 						      user_data, notify);
+	connection->blocking = FALSE;
 
 	return connection_do_open (connection, error);
 }
@@ -1435,17 +1668,20 @@ lm_connection_open_and_block (LmConnection *connection, GError **error)
 	g_return_val_if_fail (connection != NULL, FALSE);
 
 	connection->open_cb = NULL;
+	connection->blocking = TRUE;
+
 	result = connection_do_open (connection, error);
 
 	if (result == FALSE) {
 		return FALSE;
 	}
-	
+		
 	while ((state = lm_connection_get_state (connection)) == LM_CONNECTION_STATE_OPENING) {
 		if (g_main_context_pending (connection->context)) {
 			g_main_context_iteration (connection->context, TRUE);
 		} else {
-			usleep (10);
+			/* Sleep for 1 millisecond */
+			g_usleep (1000);
 		}
 	}
 
@@ -1491,7 +1727,7 @@ lm_connection_close (LmConnection      *connection,
 		     GError           **error)
 {
 	gboolean no_errors = TRUE;
-
+	
 	g_return_val_if_fail (connection != NULL, FALSE);
 
 	if (connection->state == LM_CONNECTION_STATE_CLOSED) {
@@ -1643,6 +1879,8 @@ lm_connection_authenticate_and_block (LmConnection  *connection,
 	lm_message_unref (result);
 
 	result = lm_connection_send_with_reply_and_block (connection, m, error);
+	lm_message_unref (m);
+
 	if (!result) {
 		connection->state = LM_CONNECTION_STATE_OPEN;
 		return FALSE;
@@ -1767,7 +2005,7 @@ lm_connection_set_server (LmConnection *connection, const gchar *server)
 	}
 	
 	g_free (connection->server);
-	connection->server = g_strdup (server);
+	connection->server = _lm_utils_hostname_to_punycode (server);
 }
 
 /**
@@ -2035,14 +2273,15 @@ lm_connection_send_with_reply_and_block (LmConnection  *connection,
 		lm_message_node_set_attributes (message->node, "id", id, NULL);
 	}
 
- 	g_source_remove (g_source_get_id (connection->incoming_source));
+	g_source_remove (g_source_get_id (connection->incoming_source));
 	g_source_unref (connection->incoming_source);
+	connection->incoming_source = NULL;
 
 	lm_connection_send (connection, message, error);
 
 	while (!reply) {
 		const gchar *m_id;
-		gint         n;
+		guint        n;
 
 		g_main_context_iteration (connection->context, TRUE);
 	
@@ -2190,6 +2429,7 @@ lm_connection_send_raw (LmConnection  *connection,
 			GError       **error)
 {
 	g_return_val_if_fail (connection != NULL, FALSE);
+	g_return_val_if_fail (str != NULL, FALSE);
 
 	return connection_send (connection, str, -1, error);
 }
